
--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   Analyse.hs
--
-- Author:
--   Jan Hoffmann, Klaus Aehlig (2010)
--
-- Description: 
--   To be documented.


module Analyse where

import RAMLTypes
import ATypes
import Util
import MultivariatePol
import Sharing
import Nat
import MBinomial

import qualified Data.Map as M
import Control.Monad.State.Strict
import Data.List
-- import GHC.IOBase -- useful for debugging
------------------------------------------------------
-- Typing environments,  etc
------------------------------------------------------


annotateProgram typedFunLists typedMain metric maxDegree =
    let (aTypeMain,constrMain) = annotateMain
        (aFunTypes,constrFuns) = unzip . snd $ mapAccumL (\startNames typedSCC -> runAnnotateSCC startNames typedSCC)
                                                         (tail shortNames)
                                                         typedFunLists
    in  (aTypeMain:(concat aFunTypes), constrMain++(concat constrFuns) )
    where
      annotateMain = let (aFunType, finalState) =  runState (annotateM typedMain metric maxDegree)
                                                            (startState (varNamesFrom $ head shortNames))
                     in (("main", aFunType), constraints finalState)


      runAnnotateSCC (sName:moreNames) typedSCC = 
          let fs = map (\(f,_) -> f) typedSCC
              (aFunTypes, finalState)  = runState (annotateSCC fs metric maxDegree)
                                                  (startState (varNamesFrom sName))
          in (moreNames,(aFunTypes, constraints finalState))


      varNamesFrom prefix = [ [ prefix ++ "." ++ m ++ "." ++ n | n <- shortNames ] | m <- shortNames ]

      startState names  = St { functionDefs = M.fromList (concat typedFunLists)
                             , funSCCs = makeFunSCCs M.empty typedFunLists
                             , constraints = []
                             , varNames = names
                             , cache = emptyCache
                             }

      makeFunSCCs acc [] = acc
      makeFunSCCs acc (typedSCC:typedFunLists) =
          let fs = map (\(f,_) -> f) typedSCC in
          makeFunSCCs (addSCC acc fs fs) typedFunLists

      addSCC acc0 fs scc = foldl (\acc f -> M.insert f scc acc) acc0 fs

      shortNames = [ s:xs | n <- [0..], s <- startCharacters , xs <- names n ]
        where
          names 0 = [[]]
          names n = [ x:xs | x <- baseCharacters, xs <- names (n-1)  ]

      startCharacters = ['A' .. 'Z']

      baseCharacters = ['0' .. '9'] ++ ['A' .. 'Z']




data Cache = Cache { sharingCache :: M.Map (TIndices,TIndices) (M.Map TIndices Integer)
                   , indAtMostCache :: M.Map (Int,Type) [TIndices]
                   , indDegCache :: M.Map (Int,Type) [TIndices]
                   }
      
emptyCache = Cache { sharingCache = M.empty
                   , indAtMostCache = M.empty
                   , indDegCache = M.empty
                   }

sharingCoeffCache i j =
    do { state <- get
       ; let shareCache = sharingCache (cache state)
       ; case M.lookup (i,j) shareCache of
            Just res -> return res
            Nothing ->
                do { let coeffs = sharingCoeff i j
                   ; let cache' = (cache state) {sharingCache = M.insert (i,j) coeffs shareCache }
                   ; put (state {cache = cache'} )
                   ; return coeffs
                   }
       }


indWithAtMostCache deg t =
    do { state <- get
       ; let indCache = indAtMostCache (cache state)
       ; case M.lookup (deg,t) indCache of
            Just res -> return res
            Nothing ->
                do { let inds = indicesWithDegreeAtmost deg t
                   ; let cache' = (cache state) {indAtMostCache = M.insert (deg,t) inds indCache}
                   ; put (state {cache = cache'} )
                   ; return inds
                   }
       }


indWithDegCache deg t =
    do { state <- get
       ; let indCache = indDegCache (cache state)
       ; case M.lookup (deg,t) indCache of
            Just res -> return res
            Nothing ->
                do { let inds = indicesWithDegree deg t
                   ; let cache' = (cache state) {indDegCache = M.insert (deg,t) inds indCache}
                   ; put (state {cache = cache'} )
                   ; return inds
                   }
       }



data TypeState = St { functionDefs :: FunDefs
                    , funSCCs :: M.Map FunId [FunId]
                    , constraints :: [Constraint]
                    , varNames :: [[String]]
                    , cache :: Cache
                    }

data TypeEnv = Env { currentFuns :: M.Map FunId AFunType
                   , resMetric :: ResourceMetric
                   , maxDegree :: Int
                   }


annotateM typedMain metric degree =
    do { let env = Env { currentFuns = M.empty
                       , resMetric = metric
                       , maxDegree = degree
                       }
       ; tRes <- freshAType degree (typeOf typedMain) 
       ; (AContext [] tArg) <- annotateE env typedMain tRes
       ; return $ AFunType tArg tRes
       }



annotateF f metric degree tRes =
   do { (St { funSCCs = funSCCs
            , functionDefs = funDefs }) <- get
      ; let otherFs = filter (\x -> x /= f) (funSCCs ! f)
      ; let ([(x,t1)],e_f) = funDefs ! f
      ; let t2 = typeOf e_f
      ; tArg <- freshAType degree t1
      ; let t_f = AFunType tArg tRes
      ; otherFunTypes <- sequence [freshType funDefs degree f | f <- otherFs ]
      ; let env = Env { currentFuns = M.fromList ((f,t_f):otherFunTypes)
                      , resMetric = metric
                      , maxDegree = degree
                      }
      ; _ <- sequence [annotateFunBody funDefs env f | f <- otherFunTypes]
      ; (AContext context tArgContext) <- annotateE env e_f tRes
      ; case context of
          [x] -> mapAllIndices CEq tArgContext tArg  (\i -> TITuple [i])  -- tArg == tArgContext
          []  -> constraint $ CEq [CVar (q0 tArgContext)]  [CVar (q0 tArg)]
      ; return t_f
      }


annotateSCC fs metric degree =
   do { (St { functionDefs = funDefs }) <- get
      ; funTypes <- sequence [freshType funDefs degree f | f <- fs ]
      ; let env = Env { currentFuns = M.fromList funTypes
                      , resMetric = metric
                      , maxDegree = degree
                      }
      ; sequence [annotateFunBody funDefs env f | f <- funTypes]
      }



freshType funDefs degree f = 
    do { let ([(x,t1)],e_f) = funDefs ! f
       ; let t2 = typeOf e_f
       ; tArg <- freshAType degree t1
       ; tRes <- freshAType degree t2
       ; return (f, AFunType tArg tRes)
       }

annotateFunBody funDefs env funType@(f,AFunType tArg tRes) =
    do { let ([(x,t1)],e_f) = funDefs ! f
       ; (AContext context tArgContext) <- annotateE env e_f tRes
       ; case context of
           [x] -> mapAllIndices CEq tArgContext tArg  (\i -> TITuple [i])  -- tArg == tArgContext
           []  -> constraint $ CEq [CVar (q0 tArgContext)]  [CVar (q0 tArg)]
       ; return funType
       }




freshVar :: State TypeState VarId

freshVar = do { state@(St {varNames = (var:lNames):rNames} ) <- get
              ; put (state { varNames = lNames:rNames})
              ; return var
              }


takeVars = do { state@(St {varNames = lNames:rNames} ) <- get
              ; put (state { varNames = rNames})
              ; return lNames
              }

freshAnnotation degree t = do { vNames <- takeVars
                              ; inds <- indWithAtMostCache degree t
                              ; return $ MPolynomial (M.fromList (zip inds vNames))
                              }

freshAType degree t = do { a <- freshAnnotation degree t
                                ; return (AType (t,a))
                                }


(!) q k = case M.lookup k q of
            Just r -> r
            Nothing -> error $ "Could not find key in map: " ++ (show k) ++ "\n" ++ (show q)


constraintss cs = do { state@(St {constraints = constraints}) <- get
                     ; put (state {constraints = cs ++ constraints})
                     ; return ()
                     }


constraint c = constraintss [c]


consume k t1 t2 (ResourceMetric _ metric) =
   constraint $ CGeq [CVar (q0 t1)]  [CConst (metric k), CVar (q0 t2)]


mapindices ineqsign (AType (a,MPolynomial q)) (AType (_,MPolynomial p)) f =
    let io = i0 a in
    constraintss (M.keys p >>= \ k -> if k == io then [] else [ineqsign [CVar (q ! (f k))][CVar (p ! k)]])

mapAllIndices ineqsign (AType (a,MPolynomial q)) (AType (_,MPolynomial p)) f =
    constraintss (M.keys p >>= \ k -> [ineqsign [CVar (q ! (f k))][CVar (p ! k)]])

joincontext env (AContext xs (AType ((TTuple ts),(MPolynomial p))))(AContext ys (AType ((TTuple rs),(MPolynomial q))))  cost1 cost2 =
    do
      { let vartypes = (zip xs ts) `union` (zip ys rs)
      ; let (zs,zts) = unzip vartypes
      ; agamma@(AType (_,MPolynomial qgamma)) <- freshAType (maxDegree env) (TTuple zts)
      ; let fillup (TITuple is) us = TITuple (fill vartypes (zip us is)) where
                fill []           _   = []
                fill ((x,t):xts)  uis = case lookup x uis of
                                          Just i -> i : fill xts uis
                                          Nothing -> i0 (t) : fill xts uis
      ; let makeconst env const = let (ResourceMetric _ metric) = resMetric env in CConst (metric const)
      ; constraintss (M.keys p >>= \i -> return $ CGeq [CVar (qgamma ! (fillup i xs))] 
                                                       ((CVar (p ! i)):(if i == i0 (TTuple ts) then [makeconst env cost1] else []) ) )
      ; constraintss (M.keys q >>= \i -> return $ CGeq [CVar (qgamma ! (fillup i ys))] 
                                                       ((CVar (q ! i)):(if i == i0 (TTuple rs) then [makeconst env cost2] else []) ) )
      ; return (AContext zs agamma)
      }
        


------------------------------------------------------
-- Analysing of expressions
------------------------------------------------------

annotateE :: TypeEnv -> TExp -> AType -> State TypeState AContext

annotateE env ex finalAType = case ex of
      TEShare y y' x e _ -> 
           do
             { AContext ys (AType (TTuple ts,MPolynomial pGammaEx)) <- annotateE env e finalAType
             ; let vartypes = zip ys ts
             ; let Just ty = lookup y vartypes
             ; let vartypeswithouty = filter (\(z,_) -> (z /= y) && (z /= y')) vartypes
             ; let newvartypes = (x,ty):vartypeswithouty
             ; let (newvars,newtypes) = unzip newvartypes
             ; xContext@(AType (_,MPolynomial pGammaX)) <- freshAType (maxDegree env) (TTuple newtypes) 
             ; let findys yis = fys (zip ys yis) where
                       fys [] = (error ("Sharing variable " ++ y ++ " missing"),error ("Sharing variable " ++ y' ++ " missing"),[])
                       fys ((z,iz):iys') = let  (iy,iy',is) = fys iys' in if z == y then (iz,iy',is) else if z == y' then (iy,iz,is) else (iy,iy',iz:is) 
             ; coeffs <- sequence [ do { let (i,j,is) = findys is'
                                       ; shareCoeff <- sharingCoeffCache i j
                                       ; return $ M.map (\c -> [CnVar c (pGammaEx ! (TITuple is'))]) (M.mapKeys (\ k -> TITuple (k:is)) shareCoeff)
                                       }
                                  | (TITuple is') <- M.keys pGammaEx ]
             ; let mappings = foldl (M.unionWith (++)) (M.fromList (map (\ i -> (i,[])) (M.keys pGammaX))) coeffs
             ; constraintss (M.keys pGammaX >>= \is -> [CGeq [CVar (pGammaX ! is)](mappings !is)])
             ; return $ AContext newvars xContext
             }

      TETick q _ -> 
          do 
            { a <- freshAType (maxDegree env) (TTuple [])
            ; consume (KTick q) a finalAType (resMetric env)
            ; return $ AContext [] a
            }


      TEUnit _ -> 
          do 
            { a <- freshAType (maxDegree env) (TTuple [])
            ; consume KUnit a finalAType (resMetric env)
            ; return $ AContext [] a
            }


      TETrue _ -> 
          do 
            { a <- freshAType (maxDegree env) (TTuple [])
            ; consume KBool a finalAType (resMetric env)
            ; return $ AContext [] a
            }


      TEFalse _ -> 
          do 
            { a <- freshAType (maxDegree env) (TTuple [])
            ; consume KBool a finalAType (resMetric env)
            ; return $ AContext [] a
            }


      TENat _ TInt -> 
          do 
            { a <- freshAType (maxDegree env) (TTuple [])
            ; consume KNat a finalAType (resMetric env)
            ; return $ AContext [] a
            }


      TENat n TNat ->
          do 
            { a <- freshAType (maxDegree env) (TTuple [])
            ; let AType (_,MPolynomial finalPol) = finalAType
            ; let ResourceMetric _ metric = resMetric env
            ; let mkSum (i,q) = CnVar (evalBinom (toInteger n) (toInteger $ degree i)) q
            ; let constrSum = map mkSum (M.toList finalPol)
            ; constraint $ CGeq [CVar (q0 a)]  ([CConst (metric KNat)] ++ constrSum)
            ; when (resMetric env == overflow) 
              (constraint $ CGeq [CVar (finalPol ! (TIList [TIUnit]) )] [CConst 1])
            ; return $ AContext [] a
            }
            


      TEVar x t -> -- avoid variables and inequations
          do
            { let (AType (t,MPolynomial qFinal)) = finalAType 
            ; q <- freshVar
            ; let qGamma' = M.mapKeys (\ i -> TITuple [i]) qFinal
            ; let qGamma = M.insert (TITuple [i0 t]) q qGamma'
            ; let atGamma = AType (TTuple [t],MPolynomial qGamma)
            ; consume KVar atGamma finalAType (resMetric env)
            ; return $ AContext [x] atGamma
            }


      TEApp f e t2 ->
          do
            { let (Env { resMetric = resMetric@(ResourceMetric _ metric)
                       , maxDegree = maxDegree
                       } ) = env
            ; case M.lookup f (currentFuns env) of
                Just (AFunType atArg atRes) ->
                   let (tArg,tRes) = (typeOf atArg,typeOf atRes) in
                   if maxDegree > 1 then do {    -- Add a costf-free type with degree (maxDegree-1)
                      ; freshResType <- freshAType (maxDegree - 1) tRes
                      ; (AFunType atArg' atRes') <- annotateF f costFree (maxDegree - 1) freshResType
                      ; atArgComb <- freshAType maxDegree tArg
                      ; let [pArg,pRes,pArg',pRes',pArgComb,pFinal] = map mPoly [atArg,atRes,atArg',atRes',atArgComb,finalAType]
                      ; let maybeVar i p = case M.lookup i p of
                                             Nothing -> []
                                             Just q  -> [CVar q]
                      ; constraintss (M.keys pArgComb  >>= --  tArgComb = tArg' + tArg + KApp1 
                           \i -> if i == i0 tArg then
                                     [CEq [CVar $ pArgComb ! i] (CConst (metric KApp1):(CVar $ pArg ! i):(maybeVar i pArg'))]
                                 else
                                     [CEq [CVar $ pArgComb ! i] ((CVar $ pArg ! i):(maybeVar i pArg')) ] )
                      ; constraintss (M.keys pFinal  >>=  -- finalAType + KApp2 = tRes' + tRes
                           \i -> if i == i0 tRes then
                                     [CEq [CVar $ pFinal ! i, CConst (metric KApp2)] ((CVar $ pRes ! i):(maybeVar i pRes')) ]
                                 else
                                     [CEq [CVar $ pFinal ! i] ((CVar $ pRes ! i):(maybeVar i pRes')) ] )
                      ; annotateE env e atArgComb
                      }
                      else do { -- f is current function, maxDegree == 1 , just use what we have but pass constant potential
                      ; qNewArg <- freshVar
                      ; let atArg' = AType (tArg, MPolynomial $ M.insert (i0 tArg) qNewArg (mPoly atArg) )
                      ; qConst <- freshVar
                      ; constraint $ CEq [CVar (q0 atArg')]  [CConst (metric KApp1), CVar (q0 atArg), CVar qConst]
                      ; mapindices CEq atRes finalAType id    -- XXX   avoidable?
                      ; constraint $ CEq [CVar (q0 atRes), CVar qConst ]  [CConst (metric KApp2), CVar (q0 finalAType)]
                      ; annotateE env e atArg'
                      }
                Nothing -> do { -- f is not the current function
                      ; qNewRes <- freshVar
                      ; let atRes = AType (t2, MPolynomial $ M.insert (i0 t2) qNewRes (mPoly finalAType) )
                      ; AFunType atArg _ <- annotateF f resMetric maxDegree atRes
                      ; qNewArg <- freshVar
                      ; let tArg = typeOf atArg
                      ; let atArg' = AType (tArg, MPolynomial $ M.insert (i0 tArg) qNewArg (mPoly atArg) )
                      ; consume KApp1 atArg' atArg resMetric
                      ; consume KApp2 atRes finalAType resMetric
                      ; annotateE env e atArg'
                      }
            }


      TEUnaryOp op e t -> 
          do
            { qNew <- freshVar
            ; let at = AType (t, MPolynomial $ M.insert (i0 t) qNew (mPoly finalAType))
            ; consume (KUnary op) at finalAType (resMetric env)
            ; annotateE env e at
            }


      TEBinOp Add (TEVar x1 TNat) (TEVar x2 TNat) tNat ->
         do{ qNew <- freshVar
           ; let AType (TNat, MPolynomial finalPol) = finalAType
           ; indsGamma <- indWithAtMostCache (maxDegree env) (TTuple [TNat,TNat])
           ; let gammaPol = M.fromList (map (\i@(TITuple [TIList is,TIList js]) -> (i,finalPol ! TIList (is++js))  ) indsGamma)
           ; let gammaPol' = M.insert (i0 $ TTuple [TNat,TNat]) qNew gammaPol
           ; let aGamma = AType (TTuple [TNat,TNat], MPolynomial gammaPol')
           ; consume (KBin Add) aGamma finalAType (resMetric env)
            ; when (resMetric env == overflow) 
              (constraint $ CGeq [CVar (finalPol ! (TIList [TIUnit]) )] [CConst 1])
           ; return $ AContext [x1,x2] aGamma
           }

      TEBinOp Mult (TEVar x1 TNat) (TEVar x2 TNat) tNat ->
        do{ aGamma@(AType(_,MPolynomial gammaPol)) <- freshAType (maxDegree env) (TTuple [TNat,TNat])
          ; let AType (TNat, MPolynomial finalPol) = finalAType
          ; sequence_ [ if (degree i)*2 > maxDegree env then  constraint $ CEq [CVar (finalPol M.! i)]  [CConst 0]
                           else return ()
                      | (i,q) <- M.toList finalPol ]
          ; let ResourceMetric _ metric = resMetric env
          ; let constIfZero i = if degree i == 0 then [CConst $ metric (KBin Add)] else []
          ; sequence_ [ constraint $ CGeq [CVar q_ij] $ (map (\(k,q_k) -> CnVar (multCoef i j k) q_k ) (M.toList finalPol) ) ++ (constIfZero (TITuple [i,j]) )  -- CGeq because right-hand site can be zero (issues with overflow metric)
                      | (TITuple [i,j], q_ij) <- M.toList gammaPol ]
          ; when (resMetric env == overflow) 
              (constraint $ CGeq [CVar (finalPol ! (TIList [TIUnit]) )] [CConst 1])
          ; return $ AContext [x1,x2] aGamma
          }

      TEBinOp op (TEVar x1 t1) (TEVar x2 t2) t ->
          do
            { a <- freshAType (maxDegree env) (TTuple [t1,t2])
            ; consume (KBin op) a finalAType (resMetric env)  -- no mapIndices needed since we have ground types here
            ; return $ AContext [x1,x2] a
            }

      TESucc (TEVar x TNat) TNat ->
          do 
            { let (Env { maxDegree = maxDegree
                       , resMetric = (ResourceMetric _ metric)
                       } ) = env
            ; atGamma@(AType (tGamma,MPolynomial pGamma)) <- freshAType maxDegree (TTuple [TNat])
            ; let at@(AType (_,MPolynomial pT)) = finalAType
            ; let shift (TIList l) = case M.lookup (TIList (TIUnit:l) ) pT of
                                             Nothing -> [CVar $ pT ! (TIList l) ]
                                             Just q -> [CVar q, CVar $ pT ! (TIList l)]
            ; constraintss (M.keys pGamma >>= \i@(TITuple [l]) ->
                          [CGeq [CVar (pGamma ! i) ] 
                                ((shift l) ++ (if i == i0 tGamma then [CConst (metric (KSucc))] else [] ) )
                                  ] )
            ; when (resMetric env == overflow) 
                (constraint $ CGeq [CVar (pT ! (TIList [TIUnit]) )] [CConst 1])
            ; return $ AContext [x] atGamma
            }

      TEToInt (TEVar x TNat ) TInt ->
          do
            { a <- freshAType (maxDegree env) (TTuple [TNat])
            ; consume KToInt a finalAType (resMetric env) 
            ; return $ AContext [x] a
            }

      TEMatchN (TEVar xMatch TNat) e1 xPat e2 tRes -> 
          do 
            { let (AType (_,MPolynomial pFinal)) = finalAType
            ; qNew1 <- freshVar
            ; qNew2 <- freshVar
            ; let at1 = AType (tRes, MPolynomial (M.insert (i0 tRes) qNew1 pFinal))
            ; let at2 = AType (tRes, MPolynomial (M.insert (i0 tRes) qNew2 pFinal))
            ; AContext ys1 (AType (TTuple tsG1, MPolynomial pG1)) <- annotateE env e1 at1
            ; AContext ys2 (AType (TTuple tsG2, MPolynomial pG2)) <- annotateE env e2 at2
            ; consume KMatchZ2 at1 finalAType (resMetric env)
            ; consume KMatchS2 at2 finalAType (resMetric env)
            ; let vsTsGamma = (xMatch,TNat):((zip ys1 tsG1) `union` (filter (\(y,_) -> y /= xPat) (zip ys2 tsG2)))
            ; let (ys,tsG) = unzip vsTsGamma
            ; atGamma@(AType (_,MPolynomial pGamma)) <- freshAType (maxDegree env) (TTuple tsG)
            ; let fillup (TITuple is) us vsTsGamma = TITuple (fill vsTsGamma (zip us is))      
                    where                                                                      
                      fill []           _   = []
                      fill ((x,t):xts)  uis = case lookup x uis of
                                                Just i -> i : fill xts uis
                                                Nothing -> (i0 t) : fill xts uis
            ; let (kZ1,kS1) = let (ResourceMetric _ metric) = resMetric env in 
                                      (CConst (metric KMatchZ1 )
                                      ,CConst (metric KMatchS1 ))
            ; constraintss (M.keys pG1 >>= \i -> return $ CGeq [CVar (pGamma ! (fillup i ys1 vsTsGamma))]
                                                          ( (CVar (pG1 ! i)):(if i == i0 (TTuple tsG1) then [kZ1] else []) ) )
            ; let _:vsTsGMinXm = vsTsGamma
            ; let ys2MinXs = filter (\x -> x /= xPat) ys2
            ; if xPat `elem` ys2 then do {
                ; let splitIndG2 (TITuple is) = 
                          let varInds = zip ys2 is
                              l = case lookup xPat varInds of
                                    Nothing -> []
                                    Just (TIList l') -> l'
                              is'    = snd (unzip (filter (\(x,_) -> x /= xPat) varInds))
                          in (l,TITuple is')
                ; let shift (l,i) = let TITuple is = fillup i ys2MinXs vsTsGMinXm in
                         case M.lookup (TITuple ((TIList (TIUnit:l)):is)) pGamma of
                                             Nothing -> [CVar $ pGamma ! (TITuple ((TIList l):is)) ]
                                             Just q -> [CVar q,CVar $ pGamma ! (TITuple ((TIList l):is)) ]
                ; constraintss (M.keys pG2 >>= \ind ->
                         let shiftInd = shift $ splitIndG2 ind in
                         if ind == i0 (TTuple tsG2) then
                             return $ CGeq shiftInd (kS1:[CVar (pG2 ! ind) ])
                         else
                             return $ CGeq shiftInd [CVar (pG2 ! ind) ]
                         )
                }
              else do {
                ; constraintss (M.keys pG2 >>= \i -> return $ CGeq [CVar (pGamma ! (fillup i ys2 vsTsGamma))]
                                                          ( (CVar (pG2 ! i)):(if i == i0 (TTuple tsG2) then [kS1] else []) ) )
                }
            ; return $ AContext ys atGamma
            }


      TECond (TEVar x _) e1 e2 t ->
          do
            { let (AType (_,MPolynomial pFinal)) = finalAType
            ; qNew1 <- freshVar
            ; qNew2 <- freshVar
            ; let at1 = AType (t, MPolynomial (M.insert (i0 t) qNew1 pFinal))
            ; let at2 = AType (t, MPolynomial (M.insert (i0 t) qNew2 pFinal))
            ; c1 <- annotateE env e1 at1
            ; c2 <- annotateE env e2 at2
            ; consume KConT2 at1 finalAType (resMetric env)
            ; consume KConF2 at2 finalAType (resMetric env)
            ; joincontext env c1 c2 KConT1 KConF1
            }

      TENil kind t -> 
          do 
            { a <- freshAType (maxDegree env) (TTuple [])
            ; case kind of
                 Normal -> consume KNil a finalAType (resMetric env)
                 Free   -> consume KVar a finalAType (resMetric env)  -- The freenil replaced a variable!
            ; return $ AContext [] a
            }


      TECons kind (TEVar x_h t1) (TEVar x_t t2) t ->
          do 
            { let (Env { maxDegree = maxDegree
                       , resMetric = (ResourceMetric _ metric)
                       } ) = env
            ; atGamma@(AType (tGamma,MPolynomial pGamma)) <- freshAType maxDegree (TTuple [t1,t2])
            ; let at@(AType (_,MPolynomial pT)) = finalAType
            ; let shift j (TIList l) = (case M.lookup (TIList (j:l) ) pT of
                                             Nothing -> [CConst 0]   -- XXX Change def of degree to avoid case distinction
                                             Just q -> [CVar q] )
                                       ++ (if j == i0 t1 then  [CVar $ pT ! (TIList l) ] else [])
            ; constraintss (M.keys pGamma >>= \i@(TITuple [j,l]) ->
                          [CGeq [CVar (pGamma ! i) ] 
                                ((shift j l) ++ (if kind == Normal && i == i0 tGamma 
                                                    then [CConst (metric (KCons (size t1)))] 
                                                    else [] ) )
                                  ] )
            ; return $ AContext [x_h,x_t] atGamma
            }


      TEMatchL matchType (TEVar xMatch t@(TList t1)) e1 x_h x_t e2 tRes -> 
          do 
            { let (AType (_,MPolynomial pFinal)) = finalAType
            ; qNew1 <- freshVar
            ; qNew2 <- freshVar
            ; let at1 = AType (tRes, MPolynomial (M.insert (i0 tRes) qNew1 pFinal))
            ; let at2 = AType (tRes, MPolynomial (M.insert (i0 tRes) qNew2 pFinal))
            ; AContext ys1 (AType (TTuple tsG1, MPolynomial pG1)) <- annotateE env e1 at1                             
            ; AContext ys2 (AType (TTuple tsG2, MPolynomial pG2)) <- annotateE env e2 at2
            ; consume (KMatchN2 matchType) at1 finalAType (resMetric env)
            ; consume (KMatchC2 matchType) at2 finalAType (resMetric env)
            ; let vsTsGamma = (xMatch,t):((zip ys1 tsG1) `union` (filter (\(x,_) -> not (x `elem` [x_h,x_t])) (zip ys2 tsG2)))
            ; let (ys,tsG) = unzip vsTsGamma
            ; atGamma@(AType (_,MPolynomial pGamma)) <- freshAType (maxDegree env) (TTuple tsG)
            ; let fillup (TITuple is) us vsTsGamma = TITuple (fill vsTsGamma (zip us is))      
                    where                                                                      
                      fill []           _   = []
                      fill ((x,t):xts)  uis = case lookup x uis of
                                                Just i -> i : fill xts uis
                                                Nothing -> (i0 t) : fill xts uis
            ; let (kN1,kC1) = let (ResourceMetric _ metric) = resMetric env in 
                                      (CConst (metric (KMatchN1 matchType) )
                                      ,CConst (metric (KMatchC1 matchType (size t1)) ))
            ; constraintss (M.keys pG1 >>= \i -> return $ CGeq [CVar (pGamma ! (fillup i ys1 vsTsGamma))]
                                                          ( (CVar (pG1 ! i)):(if i == i0 (TTuple tsG1) then [kN1] else []) ) )
            ; let _:vsTsGMinXm = vsTsGamma
            ; let ys2MinXs = filter (\x -> not (x `elem` [x_h,x_t]) ) ys2
            ; if (x_h `elem` ys2) || (x_t `elem` ys2) then do {
                ; let splitIndG2 (TITuple is) = 
                          let varInds = zip ys2 is
                              j = case lookup x_h varInds of
                                    Nothing -> i0 t1
                                    Just j' -> j'
                              l = case lookup x_t varInds of
                                    Nothing -> []
                                    Just (TIList l') -> l'
                              is'    = snd (unzip (filter (\(x,_) -> not (x `elem` [x_h,x_t])) varInds))
                          in (j,l,TITuple is')
                ; let shift (j,l,i) = let TITuple is = fillup i ys2MinXs vsTsGMinXm in
                         (case M.lookup (TITuple ((TIList (j:l)):is)) pGamma of
                                             Nothing -> [CConst 0]      -- XXX degree problem
                                             Just q -> [CVar q] )
                                  ++ (if j == i0 t1 then [CVar $ pGamma ! (TITuple ((TIList l):is)) ] else [])
                ; constraintss (M.keys pG2 >>= \ind ->
                         let shiftInd = shift $ splitIndG2 ind in
                         if ind == i0 (TTuple tsG2) then
                             return $ CGeq shiftInd (kC1:[CVar (pG2 ! ind) ])
                         else
                             return $ CGeq shiftInd [CVar (pG2 ! ind) ]
                         )
                }
              else do {
                ; constraintss (M.keys pG2 >>= \i -> return $ CGeq [CVar (pGamma ! (fillup i ys2 vsTsGamma))]
                                                          ( (CVar (pG2 ! i)):(if i == i0 (TTuple tsG2) then [kC1] else []) ) )
                }
            ; return $ AContext ys atGamma
            }


      TETuple es t -> 
          do
            { let varstypes = map (\(TEVar x t) -> (x,t)) es
            ; let (vars,types) = unzip varstypes
            ; qNew <- freshVar
            ; let aGamma  = AType (t, MPolynomial $ M.insert (i0 t) qNew (mPoly finalAType))
            ; consume (KTuple (size t)) aGamma finalAType (resMetric env)
            ; return $ AContext vars aGamma
            }

      TEMatchT mt (TEVar xM (TTuple tsM)) xsM e t -> 
          do  
            { let (AType (_,MPolynomial pFinal)) = finalAType
            ; newVar <- freshVar
            ; let at = (AType (t, MPolynomial (M.insert (i0 t) newVar pFinal)))
            ; AContext ysXs (atGXs@(AType (TTuple tsGXs,(MPolynomial pGXs)))) <- annotateE env e at
            ; let vsTsGammaXs = zip ysXs tsGXs
            ; let vsTsGamma = (xM,TTuple tsM):(filter (\(x,_) -> not (x `elem` xsM)) vsTsGammaXs)
            ; let (xM:ys,tsG) = unzip vsTsGamma
            ; let tG = TTuple tsG
            ; let getAnno i@(TITuple ((TITuple js):is)) = 
                      let varsInds = zip (xsM++ys) (js++is)
                          getInd (x,t) = case lookup x varsInds of
                                           Just i -> i
                                           Nothing -> i0 t
                      in
                      if all (\(x,t,j) -> j == i0 t || x `elem` ysXs) (zip3 xsM tsM js) -- index is in the context ysXs
                         then return $ (i,pGXs M.! (TITuple (map getInd vsTsGammaXs)))
                         else do { q <- freshVar ; return (i,q) }
            ; indsTsG <- indWithAtMostCache (maxDegree env) (TTuple tsG)
            ; pGammaList <- sequence [ getAnno i | i <- indsTsG ]
            ; qNew <- freshVar
            ; let pGamma = M.insert (i0 tG) qNew (M.fromList pGammaList)
            ; let context = AContext (xM:ys) (AType (tG,MPolynomial pGamma))
            ; consume (KMatchT1 mt) context atGXs (resMetric env)
            ; consume (KMatchT2 mt) at finalAType (resMetric env)
            --- error checking
--             ; let missingIndices = (indicesWithDegreeAtmost (maxDegree env) (tG) ) \\ (M.keys pGamma)
--             ; when (missingIndices /= [])
--                      (error $ "not all indices in let context.")
            ---
            ; return context
            }

      TELeaf kind _ -> 
          do 
            { a <- freshAType (maxDegree env) (TTuple [])
            ; case kind of
                 Normal -> consume KLeaf a finalAType (resMetric env)
                 Free   -> consume KLeaf a finalAType (resMetric env)  -- the freeleaf replaced a variabel
            ; return $ AContext [] a
            }

      TENodeB kind (TEVar x1 t1) (TEVar x2 t2) (TEVar x3 t3) t -> 
          do 
            { let (Env { maxDegree = maxDegree
                       , resMetric = (ResourceMetric _ metric)
                       } ) = env
            ; atGamma@(AType (tGamma,MPolynomial pGamma)) <- freshAType maxDegree (TTuple [t1,t2,t3])
            ; let at@(AType (_,MPolynomial pT)) = finalAType
            ; let shift j (TIList l1) (TIList l2) =
                      let l = l1++l2 
                          lShift = case M.lookup (TIList (j:l) ) pT of
                                     Nothing -> CConst 0   -- XXX Change def of degree to avoid case distinction
                                     Just q -> CVar q
                      in lShift:(if j == i0 t1 then [CVar $ pT ! (TIList l)] else [])
            ; constraintss (M.keys pGamma >>= \i@(TITuple [j,l1,l2]) ->
                          [CGeq [CVar (pGamma ! i) ] 
                                ((shift j l1 l2) ++ (if kind == Normal && i == i0 tGamma 
                                                     then [CConst (metric (KNodeB (size t1)))] 
                                                     else [] ) )
                                  ] )
            ; return $ AContext [x1,x2,x3] atGamma
            }          

      TEMatchB matchType (TEVar xMatch t@(TBTree t1)) e1 x1 x2 x3 e2 tRes -> 
          do 
            { let (AType (_,MPolynomial pFinal)) = finalAType
            ; qNew1 <- freshVar
            ; qNew2 <- freshVar
            ; let at1 = AType (tRes, MPolynomial (M.insert (i0 tRes) qNew1 pFinal))
            ; let at2 = AType (tRes, MPolynomial (M.insert (i0 tRes) qNew2 pFinal))
            ; AContext ys1 (AType (TTuple tsG1, MPolynomial pG1)) <- annotateE env e1 at1                             
            ; AContext ys2 (AType (TTuple tsG2, MPolynomial pG2)) <- annotateE env e2 at2
            ; consume (KMatchBL2 matchType) at1 finalAType (resMetric env)
            ; consume (KMatchBN2 matchType) at2 finalAType (resMetric env)
            ; let vsTsGamma = (xMatch,t):((zip ys1 tsG1) `union` (filter (\(x,_) -> not (x `elem` [x1,x2,x3])) (zip ys2 tsG2)))
            ; let (ys,tsG) = unzip vsTsGamma
            ; atGamma@(AType (_,MPolynomial pGamma)) <- freshAType (maxDegree env) (TTuple tsG)
            ; let fillup (TITuple is) us vsTsGamma = TITuple (fill vsTsGamma (zip us is))      
                    where                                                                      
                      fill []           _   = []
                      fill ((x,t):xts)  uis = case lookup x uis of
                                                Just i -> i : fill xts uis
                                                Nothing -> (i0 t) : fill xts uis
            ; let (kBL1,kBN1) = let (ResourceMetric _ metric) = resMetric env in 
                                      (CConst (metric (KMatchBL1 matchType) )
                                      ,CConst (metric (KMatchBN1 matchType (size t1)) ))
            ; constraintss (M.keys pG1 >>= \i -> return $ CGeq [CVar (pGamma ! (fillup i ys1 vsTsGamma))]
                                                          ( (CVar (pG1 ! i)):(if i == i0 (TTuple tsG1) then [kBL1] else []) ) )
            ; let _:vsTsGMinXm = vsTsGamma
            ; let ys2MinXs = filter (\x -> not (x `elem` [x1,x2,x3]) ) ys2

            ; if [x1,x2,x3] `intersect` ys2 /=  [] then do {
                ; let splitIndG2 (TITuple is) = 
                          let varInds = zip ys2 is
                              j = case lookup x1 varInds of
                                    Nothing -> i0 t1
                                    Just j' -> j'
                              l1 = case lookup x2 varInds of
                                    Nothing -> []
                                    Just (TIList l') -> l'
                              l2 = case lookup x3 varInds of
                                    Nothing -> []
                                    Just (TIList l') -> l'
                              is'    = snd (unzip (filter (\(x,_) -> not (x `elem` [x1,x2,x3])) varInds))
                          in (j,l1,l2,TITuple is')
                ; let shift (j,l1,l2,i) = 
                          let TITuple is = fillup i ys2MinXs vsTsGMinXm 
                              l = l1++l2
                              lShift = case M.lookup (TITuple ((TIList (j:l)):is)) pGamma of
                                         Nothing -> CConst 0      -- XXX degree problem
                                         Just q -> CVar q
                          in
                             lShift:(if j == i0 t1 then [CVar $ pGamma ! (TITuple ((TIList l):is)) ] else [])
                ; constraintss (M.keys pG2 >>= \ind ->
                         let shiftInd = shift $ splitIndG2 ind in
                         if ind == i0 (TTuple tsG2) then
                             return $ CGeq shiftInd (kBN1:[CVar (pG2 ! ind) ])
                         else
                             return $ CGeq shiftInd [CVar (pG2 ! ind) ]
                         )
                } 
              else do {
                ; constraintss (M.keys pG2 >>= \i -> return $ CGeq [CVar (pGamma ! (fillup i ys2 vsTsGamma))]
                                                          ( (CVar (pG2 ! i)):(if i == i0 (TTuple tsG2) then [kBN1] else []) ) )
                }
            ; return $ AContext ys atGamma
            } 

      TELet Normal xLet e1 e2 tFinal ->
          do
            { let (AType (_,MPolynomial pFinal)) = finalAType
            ; qNew3 <- freshVar
            ; let finalAType' = AType (tFinal, MPolynomial (M.insert (i0 tFinal) qNew3 pFinal))
            ; AContext ys2xLet at2@(AType (TTuple tsG2xLet, MPolynomial pG2xLet)) <- annotateE env e2 finalAType'
            ; consume KLet3 finalAType' finalAType (resMetric env)
            ; let (ys2,tsG2) = unzip $ filter (\(x,_) -> x /= xLet) (zip ys2xLet tsG2xLet)
            ; let tG2 = TTuple tsG2
            ; let t1 = typeOf e1
            ; if xLet `elem` ys2xLet then do {
                ; let lookupConv i (TITuple is) =
                       let indsVars = zip ys2 is
                           findInd x = 
                              if x == xLet then i
                              else let Just j = lookup x indsVars in j
                       in pG2xLet ! (TITuple $ map findInd ys2xLet)
             -- cost-free types , j \= 0
                ; let process j@(TITuple js) = do {                             
                       ; let deg = (maxDegree env) - (degree j) -- deg > 0
                       ; indices <- indWithAtMostCache deg t1
                       ; let pRes =  M.fromList $ map (\i -> (i, (lookupConv i j ) )) indices
                       ; let taRes = AType (t1,MPolynomial pRes)
                       ; let envNew = Env { resMetric = costFree
                                          , maxDegree = deg
                                          , currentFuns = M.empty  -- forget current function  (XXX sound with SCCs?)
                                          }
                       ; AContext _ (AType (_,MPolynomial pG1)) <- annotateE envNew e1 taRes
                       ; let pG1List = M.toList pG1
                       ; return $ M.fromList (map (\(TITuple is,q) -> (TITuple (is++js),q)) pG1List)
                       }
                ; indsTG2 <- indWithAtMostCache ((maxDegree env) - 1) tG2
                ; indsMaps <- sequence [ process j | j <- indsTG2 \\ [i0 tG2] ]
             -- j=0, type with current metric
                ; let j0@(TITuple j0s) = i0 tG2
                ; indices <- indWithAtMostCache (maxDegree env) t1
                ; let pRes0 =  M.fromList $ map (\i -> (i, (lookupConv i j0 ) )) indices
                ; let taRes0 = AType (t1,MPolynomial pRes0)
                ; qNew2 <- freshVar
                ; let taRes0' = AType (t1,MPolynomial (M.insert (i0 t1) qNew2 pRes0))
                ; AContext ys1 (AType (tG1@(TTuple tsG1),MPolynomial pG1_0)) <- annotateE env e1 taRes0'
                ; consume KLet2 taRes0' taRes0 (resMetric env)
                ; let pG1_0List = M.toList pG1_0
                ; let indsMap0 = M.fromList (map (\(TITuple is,q) -> (TITuple (is++j0s),q)) pG1_0List)
             -- j=maxDegree
                ; jsMax <- indWithDegCache (maxDegree env) tG2
                ; let (TITuple i0s) = i0 tG1
                ; let indsMapMax = M.fromList (map (\j@(TITuple js) -> (TITuple (i0s++js), lookupConv (i0 t1) j)) jsMax)
                ; let pGamma = foldl (\p1 p2 -> M.union p1 p2) M.empty (indsMap0:indsMapMax:indsMaps)
                ; let tG = TTuple (tsG1 ++ tsG2)
                ; let context = AContext (ys1++ys2) (AType(tG, MPolynomial pGamma))
                ; qNew1 <- freshVar
                ; let context' = AContext (ys1++ys2) (AType(tG, MPolynomial (M.insert (i0 tG) qNew1 pGamma)))
                ; consume KLet1 context' context (resMetric env)
                ; return $ context'
                }
           else do { -- xLet not in ys2
             -- j=0, type e1 with current metric
                ; let j0@(TITuple j0s) = i0 tG2
                ; taRes0 <- freshAType (maxDegree env) t1
                ; AContext ys1 (AType (tG1@(TTuple tsG1),MPolynomial pG1_0)) <- annotateE env e1 taRes0
                ; consume KLet2 taRes0 at2 (resMetric env)
                ; let pG1_0List = M.toList pG1_0
                ; let indsMap0 = M.fromList (map (\(TITuple is,q) -> (TITuple (is++j0s),q)) pG1_0List)
             -- j>0, just map indicees
                ; js <- indWithAtMostCache (maxDegree env) tG2
                ; let (TITuple i0s) = i0 tG1
                ; let indsMap = M.fromList (map (\j@(TITuple js) -> (TITuple (i0s++js), pG2xLet ! j)) js)
                ; let tG = TTuple (tsG1 ++ tsG2)
             -- all the mixed annotation between Gamma1 and Gamma2 are unconstrained 
                ; (MPolynomial pGammaDummy) <- freshAnnotation (maxDegree env) tG  -- just a dummy annotation
                ; let pGamma = M.unionWith (\q1 q2 -> q2) pGammaDummy (M.union indsMap0 indsMap) --except the unmixed
                ; let context = AContext (ys1++ys2) (AType(tG, MPolynomial pGamma))
                ; qNew1 <- freshVar
                ; let context' = AContext (ys1++ys2) (AType(tG, MPolynomial (M.insert (i0 tG) qNew1 pGamma)))
                ; consume KLet1 context' context (resMetric env)
                ; return $ context'
                }
           }

      TELet Free xLet e1 e2 _ ->
          do
            { AContext ys2xLet (AType (TTuple tsG2xLet, MPolynomial pG2xLet)) <- annotateE env e2 finalAType 
            ; let (ys2,tsG2) = unzip $ filter (\(x,_) -> x /= xLet) (zip ys2xLet tsG2xLet)
            ; let tG2 = TTuple tsG2
            ; let lookupConv i (TITuple is) =
                   let indsVars = zip ys2 is
                       findInd x = 
                          if x == xLet then i
                          else let Just j = lookup x indsVars in j
                   in pG2xLet ! (TITuple $ map findInd ys2xLet)
            ; let t1 = typeOf e1
         -- cost-free types , j \= 0
            ; let process j@(TITuple js) = do {                             
                   ; let deg = (maxDegree env) - (degree j) -- deg > 0
                   ; indices <- indWithAtMostCache deg t1
                   ; let pRes =  M.fromList $ map (\i -> (i, (lookupConv i j ) )) indices
                   ; let taRes = AType (t1,MPolynomial pRes)
                   ; let envNew = Env { resMetric = costFree
                                      , maxDegree = deg
                                      , currentFuns = M.empty -- forget current function (XXX sound with SCCs?)
                                      }
                   ; AContext _ (AType (_,MPolynomial pG1)) <- annotateE envNew e1 taRes
                   ; let pG1List = M.toList pG1
                   ; return $ M.fromList (map (\(TITuple is,q) -> (TITuple (is++js),q)) pG1List)
                   }
            ; indsTG2 <- indWithAtMostCache ((maxDegree env) - 1) tG2
            ; indsMaps <- sequence [ process j | j <- indsTG2 \\ [i0 tG2] ]
         -- j=0, type with current metric
            ; let j0@(TITuple j0s) = i0 tG2
            ; indices <- indWithAtMostCache (maxDegree env) t1
            ; let pRes0 =  M.fromList $ map (\i -> (i, (lookupConv i j0 ) )) indices
            ; let taRes0 = AType (t1,MPolynomial pRes0)
            ; AContext ys1 (AType (tG1@(TTuple tsG1),MPolynomial pG1_0)) <- annotateE env e1 taRes0
            ; let pG1_0List = M.toList pG1_0
            ; let indsMap0 = M.fromList (map (\(TITuple is,q) -> (TITuple (is++j0s),q)) pG1_0List)
         -- j=maxDegree
            ; jsMax <- indWithDegCache (maxDegree env) tG2
            ; let (TITuple i0s) = i0 tG1
            ; let indsMapMax = M.fromList (map (\j@(TITuple js) -> (TITuple (i0s++js), lookupConv (i0 t1) j)) jsMax)
            ; let pGamma = foldl (\p1 p2 -> M.union p1 p2) M.empty (indsMap0:indsMapMax:indsMaps)
--           -- error checking
--             ; let missingIndices = (indicesWithDegreeAtmost (maxDegree env) (TTuple (tsG1 ++ tsG2) )) \\ (M.keys pGamma)
--             ; when (missingIndices /= [])
--                      (error $ "not all indices in let context.")
            ; return $ AContext (ys1++ys2) (AType(TTuple (tsG1 ++ tsG2), MPolynomial pGamma))
            }

      _ -> error $ "Could not annotate the expression \n" ++ (show ex)

