--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   MemoryTransform.hs
--
-- Author:
--   Klaus Aehlig (2011)
--
-- Description:
--    For a list of function declarations make all matchings destructive
--    and all sharings calls to the clone function.

module MemoryTransform where

import RAMLTypes
import Clone
import Transform
import List
import Util

memoryvars = ["MEMTRANS_" ++ (show i) | i <- [1..]] 

clonedefinition t = (clonename t, ([("x", t)], 
                                   shareLetNormal memoryvars (clone t)))

addCloneDestructmatch :: [(String, ([(VarId, Type)], TExp))] -> [(String, ([(VarId, Type)], TExp))]
addCloneDestructmatch typedFunList = 
    (map clonedefinition (neededTypes typedFunList) ) ++ (map shareCloneDestMatch typedFunList)


neededTypes defs = nub (defs >>= gettypes >>= subtypes)
    where gettypes (_,(_,exp)) = getShareT exp


getShareT :: TExp -> [Type]

getShareT (TEShare x _ _ exp _) =  the (nub (typeLookup x exp)) : (getShareT exp)

getShareT (TEApp _ exp _) = getShareT exp
getShareT (TEUnaryOp _ exp _)  = getShareT exp
getShareT (TEBinOp _ exp1 exp2 _ ) = (getShareT exp1) ++ (getShareT exp2)
getShareT (TELet _ _ exp1 exp2 _ ) = (getShareT exp1) ++ (getShareT exp2)
getShareT (TECond exp1 exp2 exp3 _ ) = (getShareT exp1) ++ (getShareT exp2) ++ (getShareT exp3)
getShareT (TECons _ exp1 exp2 _ ) = (getShareT exp1) ++ (getShareT exp2)
getShareT (TEMatchL _ exp1 exp2  _ _ exp3 _ ) = (getShareT exp1) ++ (getShareT exp2) ++ (getShareT exp3)
getShareT (TETuple exps _) = exps >>= getShareT
getShareT (TEMatchT _ exp1 _ exp2 _ ) = (getShareT exp1) ++ (getShareT exp2)
getShareT (TENodeB _ exp1 exp2 exp3 _ ) = (getShareT exp1) ++ (getShareT exp2) ++ (getShareT exp3)
getShareT (TEMatchB _ exp1 exp2 _ _ _ exp3 _ ) = (getShareT exp1) ++ (getShareT exp2) ++ (getShareT exp3)


getShareT _ = []


typeLookup :: VarId -> TExp -> [Type]

typeLookup x (TEVar y t) = if x == y then [t] else []

typeLookup x (TEApp _ exp _) = typeLookup x exp
typeLookup x (TEUnaryOp _ exp _)  = typeLookup x exp
typeLookup x (TEBinOp _ exp1 exp2 _ ) = (typeLookup x exp1) ++ (typeLookup x exp2)
typeLookup x (TELet _ _ exp1 exp2 _ ) = (typeLookup x exp1) ++ (typeLookup x exp2)
typeLookup x (TECond exp1 exp2 exp3 _ ) = (typeLookup x exp1) ++ (typeLookup x exp2) ++ (typeLookup x exp3)
typeLookup x (TECons _ exp1 exp2 _ ) = (typeLookup x exp1) ++ (typeLookup x exp2)
typeLookup x (TEMatchL _ exp1 exp2  _ _ exp3 _ ) = (typeLookup x exp1) ++ (typeLookup x exp2) ++ (typeLookup x exp3)
typeLookup x (TETuple exps _) = exps >>= typeLookup x
typeLookup x (TEMatchT _ exp1 _ exp2 _ ) = (typeLookup x exp1) ++ (typeLookup x exp2)
typeLookup x (TENodeB _ exp1 exp2 exp3 _ ) = (typeLookup x exp1) ++ (typeLookup x exp2) ++ (typeLookup x exp3)
typeLookup x (TEMatchB _ exp1 exp2 _ _ _ exp3 _ ) = (typeLookup x exp1) ++ (typeLookup x exp2) ++ (typeLookup x exp3)

typeLookup _ _ = []


shareCloneDestMatch (f,(c_f,e_f)) = (f, (c_f, shareLetNormal memoryvars (clonematch e_f)))

clonematch :: TExp -> TExp

clonematch (TEShare x1 x2 x exp t) = let a = the (nub (typeLookup x1 exp)) in
                                     TEMatchT Normal (TEApp (clonename a) (TEVar x a) (TTuple [a,a]))
                                              [x1,x2] (clonematch exp) t

clonematch (TEMatchL _ exp1 exp2 x1 x2 exp3 t) =
    TEMatchL Dest (clonematch exp1) (clonematch exp2) x1 x2 (clonematch exp3) t

clonematch (TEMatchB _ exp1 exp2 x1 x2 x3 exp3 t) =
    TEMatchB Dest (clonematch exp1) (clonematch exp2) x1 x2 x3 (clonematch exp3) t

clonematch (TEApp f exp t) = TEApp f (clonematch exp) t
clonematch (TEUnaryOp op exp t) = TEUnaryOp op (clonematch exp) t
clonematch (TEBinOp op exp1 exp2 t) = TEBinOp op (clonematch exp1) (clonematch exp2) t
clonematch (TELet Normal x exp1 exp2 t) = TELet Normal x (clonematch exp1)(clonematch exp2) t
clonematch (TECond exp1 exp2 exp3 t) = TECond (clonematch exp1)(clonematch exp2)(clonematch exp3) t
clonematch (TECons kind exp1 exp2 t) = TECons kind (clonematch exp1)(clonematch exp2) t
clonematch (TETuple exps t) = TETuple (map clonematch exps) t
clonematch (TEMatchT mt exp1 xs exp2 t) = TEMatchT mt (clonematch exp1) xs (clonematch exp2) t
clonematch (TENodeB kind exp1 exp2 exp3 t) = TENodeB kind (clonematch exp1)(clonematch exp2)(clonematch exp3) t
clonematch (TELet Free x exp1 exp2 t) = TELet Free x (clonematch exp1)(clonematch exp2) t

clonematch exp = exp

